"""Helper to choose RGB sets of colors for figures
==================================================

Provides

.. autofunction:: make_colorchart

"""

from __future__ import division
from __future__ import print_function

# from builtins import str
from builtins import range
from colorsys import hsv_to_rgb

import matplotlib.pyplot as plt
from matplotlib.patches import Rectangle

# import subprocess
# import os
# import shlex


def make_colorchart(nb_colors=4, darkest_gray=0.03, lightest_gray=0.85):
    """Produce a pdf with color charts.

    The color charts help choosing RGB sets of colors which are
    distinguishable both in black and white and in color.

    Notes
    -----

    For many journals, it is very expensive to have figures with
    colors in the printed version. But nowadays, the printed version
    of a paper is not very important because most people get numerical
    papers on the web. So a good solution for many figures is to use
    colors that can be differentiated in black and white. This
    function produces pdf files with color charts helping researchers
    to choose colors adapted to this requirement.

    First author: Tobit CAUDWELL 2015 (tobit [AT] locta.fr).

    Parameters
    ----------

    nb_colors : int
        nb of desired final colors.

    darkest_gray : float
        darkest gray level (between 0 and 1)

    lightest_gray : float
        lightest gray level (between 0 and 1)

    """

    colors = []
    nbHue = 30  # nb of  hue values
    nbSV = 200  # nb of light/dark nuances

    # Creation of a table with colors sorted according to their hue
    for h in range(nbHue + 1):
        for sv in range(nbSV + 1):
            # variations along s
            RGBcolor = hsv_to_rgb(h / nbHue, sv / nbSV, 1)
            # multiply RGBcolor by 255
            RGBcolor = tuple(i * 255 for i in RGBcolor)

            luminance = (
                0.2126 * RGBcolor[0] + 0.7152 * RGBcolor[1] + 0.0722 * RGBcolor[2]
            )
            colors.append([h, RGBcolor, luminance])

            # variations along v
            RGBcolor = hsv_to_rgb(h / nbHue, 1, sv / nbSV)
            # multiply RGBcolor by 255
            RGBcolor = tuple(i * 255 for i in RGBcolor)

            luminance = (
                0.2126 * RGBcolor[0] + 0.7152 * RGBcolor[1] + 0.0722 * RGBcolor[2]
            )
            colors.append([h, RGBcolor, luminance])

    # round of luminance values
    lumRnd = 2.0
    nbGray = int(round(256.0 / lumRnd))
    # list of list where color are sorted
    color_table = [[0 for i in range(nbHue)] for j in range(nbGray)]
    for l in colors:
        h = l[0]
        luminance = l[2]
        # Position in color_table
        idx_lum = int(round(luminance / lumRnd)) - 1
        # First dim = luminance, 2nd dim = HUE
        color_table[idx_lum][h - 1] = l

    darkest_gray = darkest_gray * 255
    lightest_gray = lightest_gray * 255
    gray_wanted = [
        lightest_gray + i * (darkest_gray - lightest_gray) / (nb_colors - 1)
        for i in range(nb_colors)
    ]  # equivalent to linspace
    # index of desired list in color_table
    idx = [int(round(x / lumRnd)) for x in gray_wanted]

    # Selection of final colors
    final_colors = []
    for i in idx:
        colors = [x[1] for x in color_table[i]]
        final_colors.append(colors)

    #     # create the text of the latex file
    #     lines = [(
    #         r"""\documentclass{standalone}
    # \usepackage{tikz,calc}
    # \begin{document} \small
    # \begin{tikzpicture}""")]

    #     x = 0
    #     for line in reversed(final_colors):
    #         y = 0
    #         for col in line:
    #             coltxt = str([round(i/255, 2) for i in col])
    #             lines.append(
    #                 r'\definecolor{currentcolor}{rgb}{' +
    #                 coltxt[1:-1] + '}\n'
    #                 r'\node[fill=currentcolor, '
    #                 'minimum width=3cm, minimum height=1cm] at (' +
    #                 str(x) + "," + str(y) + r') {' + coltxt + '};')
    #             y -= 1
    #         x += 3.5

    #     x = 0
    #     for g in reversed(gray_wanted):
    #         coltxt = str([round(g/255, 2) for i in range(3)])
    #         lines.append(
    #             r'\definecolor{currentcolor}{rgb}{' +
    #             coltxt[1:-1] + '}\n'
    #             r'\node[fill=currentcolor, minimum width=3cm, '
    #             'minimum height=1cm] at (' +
    #             str(x) + r',2) {' + coltxt + '};')
    #         x += 3.5

    #         lines.append(r'\end{tikzpicture}' + '\n' + r'\end{document}')

    #     # write the .tex file
    #     file_name = "colorchart_using_" + str(nb_colors) + "_levels"
    #     with open(file_name + ".tex", "w") as f:
    #         f.write('\n'.join(lines))

    #     # PdfLaTeX compilation for PDF output
    #     proc = subprocess.Popen(shlex.split('pdflatex ' + file_name + ".tex"))
    #     proc.communicate()

    #     # delete .tex .log and .aux
    #     os.unlink(file_name + ".tex")
    #     os.unlink(file_name + ".log")
    #     os.unlink(file_name + ".aux")

    #     print('Document ' + file_name + '.pdf written.')

    width = 10
    height = 5
    step_x = 1.2 * width
    step_y = 1.2 * height

    fig = plt.figure(figsize=(5, 9))
    ax = fig.add_axes([0, 0, 1, 1])

    for i0, line in enumerate(reversed(final_colors)):
        for i1, color in enumerate(line):
            x = 2 + i0 * step_x
            y = 2 + i1 * step_y
            color = [round(c / 255, 2) for c in color]
            ax.add_patch(Rectangle((x, y), width, height, color=color))
            ax.text(x + 1, y + step_y / 4.0, str(color), fontsize=7)

    y += 1.5 * step_y

    for i0, g in enumerate(gray_wanted):
        color = [round(g / 255, 2)] * 3
        x = 2 + i0 * step_x
        ax.add_patch(Rectangle((x, y), width, height, color=color))
        ax.text(x + 1, y + step_y / 4.0, str(color), fontsize=7)

    ax.set_xlim([0, x + step_x])
    ax.set_ylim([0, y + step_y])
    ax.set_axis_off()

    return final_colors


if __name__ == "__main__":
    final_colors = make_colorchart(
        nb_colors=4, darkest_gray=0.5, lightest_gray=0.85
    )
    plt.show()



"""fluiddocset utility: Generate docsets compatible with Zeal and Dash
======================================================================
Docset is an easy to use offline documentation format. Different applications
such as Zeal (Linux, Windows) and Dash (Mac OS) are available to load such
docsets.  Read more about generating docsets `here
<https://kapeli.com/docsets#python>`_.

.. autofunction::  check_sphinx_build

.. autofunction::  doc_to_docsets

"""
from __future__ import print_function

import os
from shlex import split
import shutil
import argparse
from subprocess import call
from pkg_resources import resource_filename

try:
    from doc2dash.__main__ import main as doc2dash
except ModuleNotFoundError:
    print("Install doc2dash to use this tool.")

try:
    from click.testing import CliRunner
except ModuleNotFoundError:
    print("Install click to use this tool.")

from fluiddyn.io import stdout_redirected, Path


def check_sphinx_build(pkg_name, verbose, theme, regenerate=True):
    """Check if Sphinx docs has been built. If not build and return path to
    generated html.

    Parameters
    ----------
    pkg_name : str
        Name of a FluidDyn package.

    Returns
    -------
    str

    """
    doc_src = Path(resource_filename(pkg_name, "/")).parent / "doc"
    doc_build = doc_src / "_build" / "html"
    if not doc_build.exists() or regenerate:
        print("Generating Sphinx documentation ...")
        os.chdir(doc_src)
        cmd = (
            "sphinx-build -b html -d _build/doctrees "
            f"-D html_theme={theme} "
            "-D html_theme_options.nosidebar=True "
            ". _build/html"
        )
        print(cmd)
        cmd = split(cmd)
        # FIXME: with stdout_redirected(not verbose):
        call(cmd)

    return doc_src, doc_build


def doc_to_docsets(pkg_name, verbose, archive, theme):
    """Convert Sphinx docs to docsets and install them.

    Parameters
    ----------
    pkg_name : str
        Name of the Python package.

    verbose : bool
        Verbose output messages.

    archive : bool
        To make an archive or install locally.

    """
    if archive:
        doc_dest = os.getcwd()
    else:
        doc_dest = os.path.expanduser("~/.local/share/Zeal/Zeal/docsets")
        os.makedirs(doc_dest, exist_ok=True)

    doc_src, doc_build = check_sphinx_build(pkg_name, verbose, theme)
    docset_name = pkg_name.title()

    runner = CliRunner()
    cmd = "doc2dash -n {} -d {} -f -v -u https://{}.readthedocs.io ".format(
        docset_name, doc_dest, pkg_name
    )

    path_icon = doc_src / "icon.png"
    if path_icon.exists():
        cmd += f"-i {path_icon} "

    cmd += str(doc_build)
    print(cmd)
    cmd = split(cmd)
    # FIXME: with stdout_redirected(not verbose):
    result = runner.invoke(doc2dash, args=cmd[1:])

    if archive:
        doc_archive = os.path.abspath(os.path.join(doc_dest, docset_name))
        doc_format = ("gztar", ".tar.gz")
        doc_dest = os.path.join(doc_dest, docset_name + ".docset")

        # Make tarball and remove directory generated by doc2dash
        shutil.make_archive(doc_archive, doc_format[0], doc_dest)
        shutil.rmtree(doc_dest)
        doc_dest = doc_archive + doc_format[1]
    else:
        doc_dest = os.path.join(doc_dest, docset_name + ".docset")

    if verbose:
        print(result.output)

    print("Docset generated at", doc_dest)


def main():
    """Parse arguments and execute."""
    parser = argparse.ArgumentParser(
        description=__doc__, formatter_class=argparse.RawDescriptionHelpFormatter
    )

    parser.add_argument("package", help="Name of a FluidDyn package", type=str)

    parser.add_argument(
        "-a",
        "--archive",
        help="Do not install to home directory, only create a zip file.",
        action="store_true",
    )

    parser.add_argument(
        "-t",
        "--theme",
        help="a builtin or installed sphinx theme",
        default="classic",
    )
    parser.add_argument("-v", "--verbose", action="store_true")

    args = parser.parse_args()
    doc_to_docsets(args.package, args.verbose, args.archive, args.theme)


if __name__ == "__main__":
    main()



"""
IO for binary files (:mod:`fluiddyn.io.binary`)
===============================================

.. currentmodule:: fluiddyn.io.binary

Provides the class :class:`BinFile`.

.. autoclass:: BinFile
   :members:


"""

from __future__ import division, print_function


from builtins import range
import struct
import zlib
import io as _io


def chunks(l, n):
    """ Yield successive n-sized chunks from l.
    """
    for i in range(0, len(l), n):
        yield l[i : i + n]


def _code_byte_order_from_str(byteorder=None):
    if byteorder is None:
        return "="  # native

    elif byteorder.startswith("little"):
        return "<"  # little-endian

    elif byteorder.startswith("big"):
        return ">"  # big-endian

    else:
        raise ValueError("byteorder should start with little or big.")


class BinFile(_io.FileIO):

    dcodetypes = {
        "s": "s",
        "B": "B",
        "uint8": "B",
        "H": "H",
        "uint16": "H",
        "I": "I",
        "uint32": "I",
        "f": "f",
        "float32": "f",
        "d": "d",
        "float64": "d",
        "q": "q",
        "int64": "q",
    }

    keys_types = list(dcodetypes.keys())

    def __init__(self, file_path, mode="rb", byteorder=None):
        if "b" not in mode:
            mode += "b"
        super().__init__(file_path, mode)

        self.code_byte_order = _code_byte_order_from_str(byteorder)

    def readt(self, nb_values, codetype, byteorder=None):
        """Read some values coded in a particular type."""
        if codetype == "s":
            fmt = f"{nb_values:d}s"
            return struct.unpack(fmt, self.read(nb_values))[0].rstrip()

        elif codetype in self.keys_types:
            if byteorder is None:
                code_byte_order = self.code_byte_order
            else:
                code_byte_order = _code_byte_order_from_str(byteorder)

            fmt = (
                code_byte_order
                + f"{nb_values:d}"
                + self.dcodetypes[codetype]
            )
            nb_bytes = struct.calcsize(fmt)
            raw = self.read(nb_bytes)
            if len(raw) != nb_bytes:
                return "eof"  # end of file

            else:
                t_result = struct.unpack(fmt, raw)
                if len(t_result) == 1:
                    return t_result[0]

                else:
                    return t_result

        else:
            raise ValueError("Value of codetype not yet implemented")

    def readt_zlib(self, nb_bytes, nb_values, codetype):
        """Read some value encoded with zlib."""
        if codetype in self.keys_types:
            fmt = f"={nb_values:d}" + self.dcodetypes[codetype]
            nb_bytes_decompressed = struct.calcsize(fmt)
            raw = self.read(nb_bytes)
            raw_decompressed = zlib.decompress(raw)
            if len(raw_decompressed) != nb_bytes_decompressed:
                return "eof"  # end of file

            else:
                t_result = struct.unpack(fmt, raw_decompressed)
                if len(t_result) == 1:
                    return t_result[0]

                else:
                    return t_result

        else:
            raise ValueError("Value of codetype not yet implemented")

    def write_as(
        self, to_be_saved, codetype="s", byteorder=None, buffersize=1000
    ):

        if codetype == "s":
            if not isinstance(to_be_saved, bytes):
                to_be_saved = to_be_saved.encode()

            self.write(to_be_saved)
        elif codetype in self.keys_types:
            if byteorder is None:
                code_byte_order = self.code_byte_order
            else:
                code_byte_order = _code_byte_order_from_str(byteorder)

            if not hasattr(to_be_saved, "__len__"):
                fmt = code_byte_order + self.dcodetypes[codetype]
                raw = struct.pack(fmt, to_be_saved)
                self.write(raw)
            else:
                fmt_to_be_formated = (
                    code_byte_order + "{0:d}" + self.dcodetypes[codetype]
                )
                self._write_ndarray_with_buffer(
                    to_be_saved, fmt_to_be_formated, buffersize=buffersize
                )

    # raw = struct.pack(fmt, *to_be_saved)

    def _write_ndarray_with_buffer(
        self, to_be_saved, fmt_to_be_formated, buffersize=1000
    ):
        nb_values = len(to_be_saved)
        if nb_values < buffersize:
            fmt = fmt_to_be_formated.format(nb_values)
            raw = struct.pack(fmt, *to_be_saved)
            self.write(raw)
        else:
            for small_arr in chunks(to_be_saved, buffersize):
                fmt = fmt_to_be_formated.format(len(small_arr))
                raw = struct.pack(fmt, *small_arr)
                self.write(raw)


if __name__ == "__main__":
    import os
    import numpy as np

    path_test_file = os.path.expanduser("/tmp/test_file.bin")

    with BinFile(path_test_file, "w") as f:
        f.write_as("poum", buffersize=1)
        f.write_as([1, 3.55], "I", buffersize=1)
        f.write_as(np.array([1.0, 1.5]), "float64", buffersize=1)

    with BinFile(path_test_file) as f:
        s = f.readt(4, "s")
        l = f.readt(2, "I")
        a = f.readt(2, "float64")

    print(s, l, a)



"""
IO for Dantec files (:mod:`fluiddyn.io.dantec`)
===============================================

.. currentmodule:: fluiddyn.io.dantec

Provides the classes :class:`LoadedElement`, :class:`LoadedXML`
:class:`DantecImageEnsemble` and :class:`DantecVectorEnsemble`.

Warning: no security at all. For example, it would fall in a loop
if there is a loop in the file. Do not use with untrusted xml
files.

.. autoclass:: LoadedElement
   :members:

.. autoclass:: LoadedXML
   :members:

.. autoclass:: DantecImageEnsemble
   :members:

.. autoclass:: DantecVectorEnsemble
   :members:


"""

from __future__ import division, print_function

from builtins import object
import numpy as np
import os

try:
    from lxml import etree
except ImportError:
    import xml.etree.ElementTree as etree

from ast import literal_eval
from glob import glob

from fluiddyn.io.binary import BinFile


class LoadedElement:
    """A loaded element of a xml file.

    Warning: no security at all. For example, it would fall in a loop
    if there is a loop in the file. Do not use with untrusted xml
    files.

    """

    def __init__(self, unloaded_element):
        self.tag = unloaded_element.tag
        attrib = unloaded_element.attrib
        if len(attrib) > 0:
            self.attrib = attrib
        try:
            value = literal_eval(unloaded_element.text)
            self.value = value
        except (SyntaxError, ValueError):
            pass
        for child in unloaded_element:
            key = child.tag.replace(".", "_")
            keyl = "list_" + key
            # we hope there no tag in the file corresponding to keyl...
            # for Dantec files, it does not seem to be the case
            if key not in self.__dict__ and keyl not in self.__dict__:
                self.__dict__[key] = LoadedElement(child)
            else:
                if keyl not in self.__dict__:
                    self.__dict__[keyl] = []
                    self.__dict__[keyl].append(self.__dict__[key])
                    del self.__dict__[key]
                self.__dict__[keyl].append(LoadedElement(child))

    def __repr__(self):
        if hasattr(self, "attrib"):
            return repr(self.attrib)

        elif hasattr(self, "value"):
            return repr(self.value)

        else:
            return repr(self.__class__)


class LoadedXML(LoadedElement):
    """Initialize the loop on the file elements..."""

    def __init__(self, name_file):
        tree = etree.parse(name_file)
        root = tree.getroot()
        super().__init__(root)


class DantecImageEnsemble:
    _offset_header = 0xC22

    def __init__(self, path_base):
        self.path_base = path_base
        self.xlm = LoadedXML(
            os.path.join(self.path_base, "AcquiredImageEnsemble.xml")
        )
        self.name_files = glob(os.path.join(self.path_base, "data/image*.image"))
        self.shape = np.array(self.xlm.Ensemble_CoordinateSummary.imageSize.value)

    def load_image(self, ind=-1):
        name_file = self.name_files[ind]
        with BinFile(name_file) as f:
            f.seek(0)
            f.seek(self._offset_header)
            data = f.readt(self.shape.prod(), "B")
            return np.array(data, dtype=np.uint8).reshape(self.shape)


class DantecVectorEnsemble:
    def __init__(self):
        self.xlm = LoadedXML("AnalysisEnsemble_PIV.xml")


if __name__ == "__main__":

    path_base = r"/home/pa371/Data_reasons_pa371/Temp/Dantec_files/Run 11-00-21.3te2sz41/SpeedSence 1040.3te2sz4k"

    imensemble = DantecImageEnsemble(path_base)
    im = imensemble.load_image()


# vectensemble = DantecVectorEnsemble('')

# shape = np.array(imensemble.shape)
# name_file = r'image#0.image'
# f = BinFile(name_file)
# offset = 0xC22
# f.seek(offset)
# im = np.array(f.readt(shape.prod(), 'B'), dtype=np.uint8).reshape(shape)
# print(f.readt(1, 'B'))
# print(f.readt(10, 'B'))



"""Formatted xml text from xml element (:mod:`fluiddyn.util.xmltotext`)


"""

from __future__ import print_function


def produce_text_element(elem, level=0):
    lines = []
    indent_base = " " * 2
    indent = indent_base * level

    children = [child for child in elem]
    if elem.text is not None:
        text = elem.text.strip()
    else:
        text = ""
    attr_text = " ".join(
        [k + '="' + v + '"' for k, v in list(elem.attrib.items())]
    )

    if len(children) == 0 and len(text) == 0:
        if len(attr_text) > 0:
            tag_start = "<" + elem.tag + " " + attr_text + "/>"
        else:
            tag_start = "<" + elem.tag + "/>"
        tag_end = ""
    else:
        if len(attr_text) > 0:
            tag_start = "<" + elem.tag + " " + attr_text + ">"
        else:
            tag_start = "<" + elem.tag + ">"
        tag_end = "</" + elem.tag + ">"

    if len(children) > 0 or "\n" in text or len(text) > 30:
        # "long" element
        lines.append(format_too_long_tagstart(indent + tag_start))
        if len(text) > 0:
            lines.append(format_too_long_text(indent + indent_base + text))
        for child in children:
            lines.append(produce_text_element(child, level=level + 1))
        lines.append(indent + tag_end)
    else:
        lines.append(
            format_too_long_tagstart(indent + tag_start)
            + " "
            + text
            + " "
            + tag_end
        )

    txt = "\n".join(lines) + "\n"
    try:
        txt = txt.decode("UTF-8")
    except (AttributeError, UnicodeEncodeError):
        pass

    return txt


def get_position_first_letter(text):
    i = 0
    pos_space = text.find(" ", i)
    while i == pos_space:
        i += 1
        pos_space = text.find(" ", i)
    return i


def format_too_long_text(text, lengthmax=79):
    if len(text) <= lengthmax:
        return text

    lines = []
    position_first_letter = get_position_first_letter(text)
    indent = " " * position_first_letter
    line = indent
    for word in text.split(" "):
        if len(line + word) > lengthmax:
            lines.append(line)
            line = indent + word
        else:
            if line.endswith(" "):
                line = line + word
            else:
                line = line + " " + word
    lines.append(line)

    return "\n".join(lines)


def get_indent_after_tag(text):
    i = 0
    pos_space = text.find(" ", i)
    while i == pos_space:
        i += 1
        pos_space = text.find(" ", i)
    if pos_space == -1:
        nb_space_indent = 0
    else:
        nb_space_indent = pos_space + 1
    return " " * nb_space_indent


def format_too_long_tagstart(text, lengthmax=79):
    if len(text) <= lengthmax:
        return text

    lines = []
    words = text.split(" ")

    position_first_letter = get_position_first_letter(text)
    line = " " * position_first_letter + words[0]
    indent = get_indent_after_tag(text)

    for word in words[1:]:
        if len(line + word) > lengthmax:
            lines.append(line)
            line = indent + word
        else:
            if line.endswith(" "):
                line = line + word
            else:
                line = line + " " + word
    lines.append(line)

    return "\n".join(lines)


if __name__ == "__main__":

    try:
        from lxml import etree
    except ImportError:
        import xml.etree.ElementTree as etree

    xmltext = """
<data>
    <country name="Liechtenstein">
        <rank updated="yes">2</rank>
        <year>2008</year>
        <gdppc>141100</gdppc>
        <neighbor name="Austria" direction="E" attr1="long long long" attr2="long long long" attr3="long long long">

    blablabla bliblibli bloublou blublu bloblo blablabla bliblibli bloublou blublu bloblo blablabla bliblibli bloublou blublu bloblo

        </neighbor>

        <neighbor name="Switzerland" direction="W"/>
    </country>

    <country name="Singapore">

 <rank updated="yes">5 and a text which can actually be very very long something quite long long long long long.</rank>

        <year>2011</year>
        <gdppc>59900</gdppc>
        <neighbor name="Malaysia" direction="N"/>
    </country>
    <country name="Panama">
        <rank updated="yes">69</rank>
        <year>2011</year>
        <gdppc>13600</gdppc>
        <neighbor name="Costa Rica" direction="W"/>
        <neighbor name="Colombia" direction="E"/>
    </country>
</data>

"""

    elem = etree.fromstring(xmltext)

    print(produce_text_element(elem))



"""
FluidDyn
========

Provides 

  1. An object-oriented and modular set of solvers.
  2. Tools useful for carrying out experiments in fluid dynamics.

The docstring examples assume that `fluiddyn` has been imported as `fld`::

  >>> import fluiddyn as fld

Use the question mark in ipython to view a function's docstring::

  >>> fld.create_object_from_file?
  ... # doctest: +SKIP

"""

from .util.util import (
    create_object_from_file,
    time_as_str,
    get_memory_usage,
    ipydebug,
)
from .util import constants
from ._version import __version__

__all__ = [
    "__version__",
    "constants",
    "create_object_from_file",
    "time_as_str",
    "get_memory_usage",
    "ipydebug",
]



"""
IO for HDF5 files (:mod:`fluiddyn.io.hdf5`)
===========================================

.. autoclass:: H5File
   :members:

.. autofunction:: save_variables_h5

.. autofunction:: load_variables_h5

"""


from __future__ import division, print_function

from builtins import range
import numpy as np
import numbers
import h5py


class H5File(h5py.File):
    """HDF5 file."""

    def save_dict(self, keydict, dicttosave):
        """Save the dictionnary `dicttosave` in the file.

        """
        group_params = self.create_group(keydict)
        if len(dicttosave) > 0:
            for k, v in list(dicttosave.items()):
                group_params.create_dataset(k, data=v)

    def load_dict(self, keydict):
        """Load a group as a dictionnary.

        """
        group_params = self[keydict]
        params = {}
        for k, v_in_file in list(group_params.items()):
            v = v_in_file[...]
            if not isinstance(v, (float, int)) and v.ndim == 0:
                v = v.item()
            params[k] = v
        return params

    def update_dict(self, keydict, dicttosave):
        group_params = self[keydict]
        for k, v in list(dicttosave.items()):
            group_params.create_dataset(k, data=v)

    def save_dict_of_ndarrays(self, dicttosave, dtype=np.float32):
        """Save ndarrays in the file."""

        dicttosave1 = {}
        for k, v in list(dicttosave.items()):
            if isinstance(v, numbers.Number):
                v = [v]
            a = np.array(v, dtype=dtype)
            dicttosave1[k] = a[None, ...]

        if k not in list(self.keys()):
            for k, v in list(dicttosave1.items()):
                self.create_dataset(k, data=v, maxshape=(None,) + v.shape[1:])
        else:
            nb_saved_times = self[k].shape[0]
            for k, v in list(dicttosave1.items()):
                dset_p = self[k]
                dset_p.resize((nb_saved_times + 1,) + v.shape[1:])
                dset_p[nb_saved_times] = v

    def load(self, times_slice=None):
        """Load data."""
        dict_return = {}
        for k in list(self.attrs.keys()):
            dict_return[k] = self.attrs[k]

        for k in list(self.keys()):
            dict_return[k] = self[k][...]

        if times_slice is not None:
            try:
                times = dict_return["times"]
            except KeyError:
                raise ValueError("No array times in the file.")

            tstart = times_slice[0]

            if tstart is None:
                tstart = times[0]
            itstart = abs(times - tstart).argmin()

            if len(times_slice) > 1 and times_slice[1] is not None:
                tend = times_slice[1]
            else:
                tend = times[-1]
            itend = abs(times - tend).argmin()

            if len(times_slice) > 2 and times_slice[2] is not None:
                tstep = times_slice[2]
            else:
                tstep = 0.0

            its = [itstart]
            for it in range(itstart + 1, itend):
                if times[it] >= times[its[-1]] + tstep:
                    its.append(it)
            its = list(its)

            for k in list(self.keys()):
                dict_return[k] = dict_return[k][its]

        return dict_return


def save_variables_h5(path, variables, names=None):
    """Save data in `variables` in the file `path`.

    Parameters
    ----------

    path : str

      Path of the file where the data are saved (has to end with '.py').

    variables : dict

      Contains the variables to be saved.

    names : None or sequence of str.

      If None, all variables in variables are saved, else, only the variables
      with name in names of the variables to be saved.

    Examples
    --------

    .. code::

       a = 1
       b = 'str'
       c = np.ones(2)

       save_variables_h5('myfile.py', locals(), ('a', 'b', 'c'))

    or from a dictionary::

       d = {'a': 1, 'b': 'str', 'c': np.ones(2)}
       save_variables_h5('myfile.py', d)

    """
    if names is not None:
        variables_all = variables
        variables = {}
        for name in names:
            variables[name] = variables_all[name]

    with h5py.File(path, "w") as f:
        for key, value in variables.items():
            f.create_dataset(key, data=value)


def load_variables_h5(path):
    """Load files created with the function `save_variables_h5`.

    Parameters
    ----------

    path : str

      Path towards a hdf5 file saved with `save_variables_h5`.

    """

    variables = {}

    with h5py.File(path, "r") as f:
        for key, dataset in f.items():
            variables[key] = dataset.value

    return variables


# if __name__ == '__main__':

#     pass



"""Clean Matlab code
====================

"""

from __future__ import print_function

import os

from . import get_index_closing_parenthesis

block_definers = ("if", "while", "for")

ident = 4 * " "


def is_comment_line(line):
    return line.strip().startswith("%")


def modif_identation(code_lines):

    ident_level = 0
    lines_new = []

    for line in code_lines:
        line = line.strip()
        if line == "":
            lines_new.append(line)
            continue

        ident_direction_for_next = 0
        ident_this_line = 0

        # first_name = line.split(maxsplit=1)[0]
        # for Python 2.7
        first_name = line.split(None, 1)[0]
        if first_name in block_definers:
            ident_direction_for_next = 1
        elif first_name in ("end", "end;"):
            ident_direction_for_next = -1
            ident_this_line = -1
        elif first_name == "elseif":
            ident_direction_for_next = 0
            ident_this_line = -1

        lines_new.append((ident_level + ident_this_line) * ident + line)
        ident_level += ident_direction_for_next

    return lines_new


def modif_unwrap_if(code_lines):
    lines_new = []
    for line in code_lines:
        line = line.strip()
        if_starts = ("if (", "elseif (")

        if not any(line.startswith(s) for s in if_starts):
            lines_new.append(line)
            continue

        for s in if_starts:
            if line.startswith(s):
                index_begin = len(s) - 1

        index = get_index_closing_parenthesis(line, index_begin)
        if_statement = line[: index + 1]
        lines_new.append(if_statement)
        rest_line = line[index + 1 :]

        if len(rest_line) == 0:
            continue

        lines_new.append(rest_line)

    return lines_new


def modif_split_statements(code_lines):
    lines_new = []
    for line in code_lines:
        line = line.strip()

        if is_comment_line(line) or ";" not in line:
            lines_new.append(line)
            continue

        statements = [stat.strip() for stat in line.split(";")]

        for i, stat in enumerate(statements[:-1]):
            statements[i] = stat + ";"

        if statements[-1] == "":
            statements = statements[:-1]

        lines_new.extend(statements)

    return lines_new


def modif_split_comments_from_code(code_lines):
    lines_new = []
    for line in code_lines:
        line = line.strip()

        if is_comment_line(line) or "%" not in line:
            lines_new.append(line)
            continue

        parts = line.split("%")
        before_comment = ""
        for i, part in enumerate(parts):
            before_comment += part
            if not before_comment.endswith("\\"):
                break

            else:
                before_comment += "%"

        comment_parts = parts[i + 1 :]
        if len(comment_parts) == 0:
            # no comment in the line
            lines_new.append(line)
            continue

        comment = "%" + "%".join(comment_parts)

        if line.startswith("if "):
            lines_new.extend([before_comment, comment])
        else:
            lines_new.extend([comment, before_comment])

    return lines_new


def modif_spaces_around_operators(code_lines):
    lines_new = []
    for line in code_lines:

        if is_comment_line(line) or "=" not in line or "for " in line:
            lines_new.append(line)
            continue

        parts = line.split("=")

        new_line = ""

        for i, part in enumerate(parts[:-1]):
            new_line += part

            if part != "" and not part.endswith(" "):
                new_line += " "

            new_line += "="

            if parts[i + 1] != "" and not parts[i + 1].startswith(" "):
                new_line += " "

        new_line += parts[-1]
        lines_new.append(new_line)

    return lines_new


def modif_space_after_comma(code_lines):
    lines_new = []
    for line in code_lines:

        if is_comment_line(line):
            lines_new.append(line)
            continue

        parts = line.split(",")

        new_line = ""
        for i, part in enumerate(parts[:-1]):
            new_line += part + ","

            if not parts[i + 1].startswith(" "):
                new_line += " "

        new_line += parts[-1]
        lines_new.append(new_line)

    return lines_new


def modif_code(path_file):

    with open(path_file) as f:
        code_lines = f.readlines()

    modif_functions = [
        modif_split_comments_from_code,
        modif_split_statements,
        modif_unwrap_if,
        modif_spaces_around_operators,
        modif_space_after_comma,
        modif_identation,
    ]

    for func in modif_functions:
        code_lines = func(code_lines)

    new_code = "\n".join(code_lines)
    return new_code


def treat_matlab_directory(path_dir):

    path_cleaner = path_dir + "_cleaner"

    if not os.path.exists(path_cleaner):
        os.mkdir(path_cleaner)

    for root, subdirs, nfiles in os.walk(path_dir):
        relative_path = root[len(path_dir) :]

        new_dir = path_cleaner + relative_path

        if not os.path.exists(new_dir):
            os.mkdir(new_dir)

        matlab_nfiles = [nfile for nfile in nfiles if nfile.endswith(".m")]

        for nfile in matlab_nfiles:
            path_file = os.path.join(root, nfile)
            path_file_new = os.path.join(new_dir, nfile)

            print((path_file, path_file_new))

            new_code = modif_code(path_file)

            with open(path_file_new, "w") as f:
                f.write(new_code)


if __name__ == "__main__":

    path_dir = "diablo_mat"

    treat_matlab_directory(path_dir)

# code = modif_code(path_dir + '/diablo.m')
# print(code)

# code_lines = [
#     "    TIME=TIME+DELTA_T;",
#     "    if (mod(TIME_STEP,10)==0 & toto == 2)"]
# print('\n'.join(modif_spaces_around_operators(code_lines)))



from builtins import range
import unittest
import os
from shutil import rmtree

import numpy as np
from PIL import Image


from fluiddyn.util.serieofarrays import SeriesOfArrays, SerieOfArraysFromFiles


def create_image(path):
    im = Image.fromarray(np.ones((8, 8), dtype=np.int32))
    im.save(path)
    im.close()


class TestSeriesOfArrays(unittest.TestCase):
    @classmethod
    def setUpClass(cls):
        work_dir = cls.work_dir = "test_fluiddyn_util_serieofarrays"
        if not os.path.exists(work_dir):
            os.makedirs(work_dir)
        for i in range(7):
            for j in range(2):
                create_image(os.path.join(work_dir, f"file{i}_{j}.png"))

    @classmethod
    def tearDownClass(cls):
        rmtree(cls.work_dir)

    def test0(self):

        serie = SerieOfArraysFromFiles(self.work_dir, "1:1+3, :")
        serie.get_path_arrays()
        serie.get_name_path_arrays()
        serie.check_all_arrays_exist()

        serie = SerieOfArraysFromFiles(self.work_dir, "1:1+3:2, 1")
        assert len(serie.get_name_files()) == 2

        series = SeriesOfArrays(
            self.work_dir, "i:i+3:2, 1", ind_start=0, ind_stop=None, ind_step=2
        )

        series = SeriesOfArrays(
            self.work_dir, "i:i+3:2, 1", ind_start=0, ind_stop=8, ind_step=2
        )

        self.assertEqual(len(series), 4)
        self.assertEqual(series.ind_stop, 7)
        series.get_next_serie()
        series.get_name_all_files()
        series.get_name_all_arrays()

        tuple(series.items())

        serie = series.get_serie_from_index(0)
        serie.get_arrays()
        serie.get_array_from_index(0)
        serie.get_array_from_indices(0, 1)
        serie.get_name_files()
        serie.get_path_all_files()
        serie.get_path_files()

        for path in serie.iter_path_files():
            pass

        serie.get_index_slices_all_files()
        serie.get_index_slices()
        serie.get_nb_files()
        serie.set_index_slices(0, 1)


if __name__ == "__main__":
    unittest.main()



"""
Slurm clusters under CINES (:mod:`fluiddyn.clusters.cines`)
===========================================================

Provides:

.. autoclass:: Occigen
   :members:

"""

from fluiddyn.clusters.slurm import ClusterSlurm


class Occigen(ClusterSlurm):
    name_cluster = "occigen"
    nb_cores_per_node = 24
    constraint = "HSW24"
    max_walltime = "23:59:59"
    cmd_run = "srun --mpi=pmi2 -K1 --resv-ports"

    def __init__(self):
        super().__init__()
        self.check_name_cluster()

        self.commands_setting_env = [
            # 'OPT=/scratch/cnt0022/egi2153/SHARED/opt',
            "OPT=$HOME/opt",
            "module purge",
            "module load gcc/6.2.0",
            "module load intel/17.2",
            "module load openmpi/intel/2.0.2",
            "module load qt",
            "module load hdf5-seq",
            "module load python/3.6.3",
            "unset PYTHONPATH",
            "source $HOME/mypy/bin/activate",
            (
                "export LD_LIBRARY_PATH=$LD_LIBRARY_PATH:"
                "$OPT/pfft/lib:$OPT/p3dfft/2.7.5/lib:$OPT/fftw/3.3.7/lib"
            ),
            "export LANG=C",
            "export LC_LANG=C",
        ]


class Occigen28(Occigen):
    nb_cores_per_node = 28
    constraint = "BDW28"



import unittest
import os
import numpy as np

from ...io import stdout_redirected
from .. import easypyfft

try:
    import pulp

    pulp_import_error = False
except ImportError:
    pulp_import_error = True


class TestFFTWGrid(unittest.TestCase):
    def tearDown(self):
        lpfile = "FFTWGridSizeOptimizationModel.lp"
        if os.path.exists(lpfile):
            os.remove(lpfile)

    @unittest.skipIf(
        pulp_import_error,
        "fftw_grid_size depends on the package pulp, which is not installed.",
    )
    def test_fftw_grid_size(self):
        with stdout_redirected():
            n = easypyfft.fftw_grid_size(1020, debug=True)

        self.assertEqual(n, 1024)


class TestFFTW1D(unittest.TestCase):
    def test_fft(self):
        op = easypyfft.FFTW1D(4)
        func_fft = np.zeros(op.shapeK, dtype=np.complex128)
        func = op.ifft(func_fft)
        op.fft(func)


class TestFFTW1DReal2Complex(unittest.TestCase):
    def test_fft(self):
        """Should be able to..."""
        nx = 64
        op = easypyfft.FFTW1DReal2Complex(nx)

        func_fft = np.zeros(op.shapeK, dtype=np.complex128)
        func_fft[0] = 1

        self.compute_and_check(func_fft, op)

    def compute_and_check(self, func_fft, op):

        func = op.ifft(func_fft)
        back_fft = op.fft(func)
        back = op.ifft(back_fft)

        self.assertTrue(np.allclose(func_fft, back_fft))
        self.assertTrue(np.allclose(func, back))

        energyX = op.compute_energy_from_spatial(func)
        energyK = op.compute_energy_from_Fourier(func_fft)
        energyKback = op.compute_energy_from_Fourier(back_fft)

        self.assertAlmostEqual(energyX, energyK)
        self.assertAlmostEqual(energyK, energyKback)

    def test_fft_random(self):
        """Should be able to..."""
        nx = 64
        op = easypyfft.FFTW1DReal2Complex(nx)

        func_fft = np.random.random(op.shapeK) + 1.0j * np.random.random(
            op.shapeK
        )
        func = op.ifft(func_fft)
        func_fft = op.fft(func)

        self.compute_and_check(func_fft, op)


class TestFFTW2DReal2Complex(unittest.TestCase):
    cls = easypyfft.FFTW2DReal2Complex

    def test_fft(self):
        """Should be able to..."""
        nx = 4
        ny = 2
        op = self.cls(nx, ny)

        func_fft = np.zeros(op.shapeK, dtype=np.complex128)
        func_fft[0, 1] = 1

        with stdout_redirected():
            op.run_tests()
            op.run_benchs(1)

        self.compute_and_check(func_fft, op)

    def compute_and_check(self, func_fft, op):

        energyK = op.compute_energy_from_Fourier(func_fft)

        func = op.ifft2d(func_fft)
        energyX = op.compute_energy_from_spatial(func)

        back_fft = op.fft2d(func)
        energyKback = op.compute_energy_from_Fourier(back_fft)
        back = op.ifft2d(back_fft)

        self.assertTrue(np.allclose(func_fft, back_fft))
        self.assertTrue(np.allclose(func, back))

        self.assertAlmostEqual(energyX, energyK)
        self.assertAlmostEqual(energyK, energyKback)

    def test_fft_random(self):
        """Should be able to..."""
        nx = 16
        ny = 32
        op = self.cls(nx, ny)

        func_fft = np.random.random(op.shapeK) + 1.0j * np.random.random(
            op.shapeK
        )

        with stdout_redirected():
            func = op.ifft2d(func_fft)
        func_fft = op.fft2d(func)

        self.compute_and_check(func_fft, op)


class TestFFTP2D(TestFFTW2DReal2Complex):
    cls = easypyfft.FFTP2D


class TestFFTW3DReal2Complex(unittest.TestCase):
    def test_fft(self):
        """Test easypyfft.FFTW3DReal2Complex"""
        nx = 4
        ny = 2
        nz = 8
        op = easypyfft.FFTW3DReal2Complex(nx, ny, nz)

        func_fft = np.zeros(op.shapeK, dtype=np.complex128)
        func_fft[0, 0, 1] = 1

        self.compute_and_check(func_fft, op)

        op.get_shapeX_loc()
        nX0, nX1, nX2 = op.get_shapeX_seq()
        op.get_shapeK_loc()
        nK0, nK1, nK2 = op.get_shapeK_seq()
        op.get_k_adim()
        op.get_k_adim_loc()
        op.get_dimX_K()
        op.get_seq_indices_first_K()

        o2d = easypyfft.FFTW2DReal2Complex(nX2, nX1)

        arr2d = np.empty(o2d.get_shapeX_loc())
        arr2d_fft = o2d.fft2d(arr2d)
        op.build_invariant_arrayX_from_2d_indices12X(o2d, arr2d)
        op.build_invariant_arrayK_from_2d_indices12X(o2d, arr2d_fft)

        op.create_arrayX(1)
        op.create_arrayK(1)

    def compute_and_check(self, func_fft, op):

        energyK = op.compute_energy_from_Fourier(func_fft)

        func = op.ifft(func_fft)
        energyX = op.compute_energy_from_spatial(func)

        back_fft = op.fft(func)
        energyKback = op.compute_energy_from_Fourier(back_fft)
        back = op.ifft(back_fft)

        self.assertTrue(np.allclose(func_fft, back_fft))
        self.assertTrue(np.allclose(func, back))

        self.assertAlmostEqual(energyX, energyK)
        self.assertAlmostEqual(energyK, energyKback)

    def test_fft_random(self):
        """Should be able to..."""
        nx = 2
        ny = 4
        nz = 8
        op = easypyfft.FFTW3DReal2Complex(nx, ny, nz)

        func_fft = np.random.random(op.shapeK) + 1.0j * np.random.random(
            op.shapeK
        )
        func = op.ifft(func_fft)
        func_fft_back = op.fft(func)

        self.compute_and_check(func_fft_back, op)


if __name__ == "__main__":
    unittest.main()



"""dump file utility (:mod:`fluiddyn.io.dump`)
==============================================

"""

from __future__ import print_function

import os
from glob import glob
import argparse

import h5py


description = """

Utility to print the hierarchy of hdf5 and netcdf files.

"""

__doc__ += description


def dump(path, print_variables=False):
    print("dump file:", path)
    if path.endswith(".nc"):
        dump_nc_file(path, print_variables=print_variables)
    elif path.endswith(".h5"):
        dump_h5_file(path, print_variables=print_variables)


def dump_nc_file(path, print_variables=False):
    import h5netcdf

    with h5netcdf.File(path) as obj:
        dump_nc_object(obj, level=0, print_variables=print_variables)


def dump_nc_object(obj, level=0, print_variables=False):
    indent1 = "  "
    indent = level * indent1
    indentp1 = indent + indent1

    print(indent + "netcdf object: " + obj.name)

    if len(obj.attrs) != 0:
        print(indentp1 + "attrs:", dict(obj.attrs))

    if len(obj.dimensions) != 0:
        print(indentp1 + "dimensions:", dict(obj.dimensions))

    if len(obj.variables) > 0:
        print(indentp1 + "variables:")
        for value in obj.variables.values():
            print(
                indentp1
                + "- variable {} (dimensions {}, dtype {})".format(
                    value.name, value.dimensions, value.dtype
                )
            )
            if len(value.attrs) != 0:
                print(indent1 + indentp1 + "attrs:", dict(value.attrs))
            if print_variables:
                print(indent1 + indentp1 + repr(value[:]))

    if len(obj.groups) > 0:
        print(indentp1 + "groups:")
        for name_group in obj.groups:
            print(indentp1 + "- " + name_group)

        for group in obj.groups.values():
            print()
            dump_nc_object(
                group, level=level + 1, print_variables=print_variables
            )


def dump_h5_file(path, print_variables=False):
    with h5py.File(path, "r") as obj:
        dump_h5_object(obj, level=0, print_variables=print_variables)


def dump_h5_object(obj, level=0, print_variables=False):
    indent1 = "  "
    indent = level * indent1
    indentp1 = indent + indent1

    print(indent + "hdf5 object: " + obj.name)

    if len(obj.attrs) != 0:
        print(indentp1 + "attrs:", dict(obj.attrs))

    variables = []
    groups = []
    for tmp in obj.values():
        try:
            tmp.keys
        except AttributeError:
            variables.append(tmp)
        else:
            groups.append(tmp)

    if len(variables) > 0:
        print(indentp1 + "variables:")
        for value in variables:
            print(
                indentp1
                + f"- variable {value.name} (dtype {value.dtype})"
            )
            if len(value.attrs) != 0:
                print(indent1 + indentp1 + "attrs:", dict(value.attrs))
            if print_variables:
                print(indent1 + indentp1 + repr(value[:]))

    if len(groups) > 0:
        print(indentp1 + "groups:")
        for group in groups:
            print(indentp1 + "- " + group.name)

        for group in groups:
            print()
            dump_h5_object(
                group, level=level + 1, print_variables=print_variables
            )


def main():

    parser = argparse.ArgumentParser(
        prog="fluiddump",
        description=description,
        formatter_class=argparse.RawDescriptionHelpFormatter,
    )

    parser.add_argument(
        "file", help="str indicating which file has to be dump.", type=str
    )

    parser.add_argument(
        "-pv",
        "--print-variables",
        help="also print the content of the variables",
        action="store_true",
    )

    args = parser.parse_args()

    if os.path.isfile(args.file):
        dump(args.file, print_variables=args.print_variables)
    else:
        paths = glob(args.file)
        if len(paths) == 0:
            raise ValueError("No file found from the input string:\n" + args.file)

        for path in paths:
            dump(path, print_variables=args.print_variables)



"""
Test multitiff module
=====================

"""

import unittest
import os
from shutil import rmtree

import numpy as np
from PIL import Image

from ..multitiff import (
    imsave,
    reorganize_single_frame_3Dscannedpiv_data,
    reorganize_piv3dscanning_doubleframe,
    reorganize_piv2d_singleframe,
    reorganize_piv2d_doubleframe,
)

from ..redirect_stdout import stdout_redirected


class TestMultiTIFF(unittest.TestCase):
    """Test fluiddyn.io.multitiff module."""

    @classmethod
    def setUpClass(cls):
        def im(maxval, dtype, shape=(10, 10)):
            """Generate a test image."""
            return (maxval * np.random.random(shape)).astype(dtype)

        cls._work_dir = "test_fluiddyn_io_multitiff"
        if not os.path.exists(cls._work_dir):
            os.mkdir(cls._work_dir)

        os.chdir(cls._work_dir)
        cls.n_frames = 4
        cls.nb_files = 2

        for ifile in range(cls.nb_files):
            arrays = [im(2 ** 8 - 1, np.int8) for i in range(cls.n_frames)]
            imsave(f"test_multitiff{ifile}.tif", arrays, as_int=True)

    @classmethod
    def tearDownClass(cls):
        os.chdir("..")
        rmtree(cls._work_dir)

    def test_imsave(self):
        with Image.open("test_multitiff0.tif") as im:
            n_frames = im.n_frames

        if n_frames != self.n_frames:
            raise ValueError(
                "Multiframe TIFF imsave unsuccessful. No. of frames={}".format(
                    n_frames
                )
            )

    def test_single_frame_3Dscannedpiv(self):
        with stdout_redirected():
            reorganize_single_frame_3Dscannedpiv_data(
                "test*.tif", 2, outputdir=".", outputext="png", erase=True
            )

    def test_piv3dscanning_doubleframe(self):
        with stdout_redirected():
            reorganize_piv3dscanning_doubleframe(
                "test*.tif", 2, outputdir=".", outputext="png", erase=False
            )

    def test_piv2d_singleframe(self):
        with stdout_redirected():
            reorganize_piv2d_singleframe(
                "test*.tif", outputdir=".", outputext="png", erase=False
            )

    def test_piv2d_doubleframe(self):
        with stdout_redirected():
            reorganize_piv2d_doubleframe(
                "test*.tif", outputdir=".", outputext="png", erase=False
            )


if __name__ == "__main__":
    unittest.main()



"""
Test SLURM and SNIC clusters
============================

"""
import unittest
import os
from shutil import rmtree
from .. import pbs
from ...io import stdout_redirected


jobid = 123


def my_test_input(message):
    return f"{jobid} {jobid}"


pbs.input = my_test_input


class ClusterPBSMod(pbs.ClusterPBS):
    """A modified class which skips checking if PBS is installed or not."""

    cmd_launch = "echo"

    def check_pbs(self):
        pass

    def check_name_cluster(self, env="HOSTNAME"):
        os.environ[env] = self.name_cluster
        super().check_name_cluster(env)


class PBSTestCase(unittest.TestCase):
    """Test ClusterPBS submit_script method."""

    def setUp(self, cls=ClusterPBSMod):
        self._work_dir = "test_pbs"
        self._script = "script.py"
        self._script_resume = "script_resume.py"
        if not os.path.exists(self._work_dir):
            os.mkdir(self._work_dir)

        os.chdir(self._work_dir)
        with open(self._script, "w") as f:
            f.write('print("Hello world")')

        with open(self._script_resume, "w") as f:
            f.write('print("Hello Again")')

        self.cluster = cls()

    def tearDown(self):
        os.chdir("..")
        try:
            rmtree(self._work_dir)
        except OSError:
            # in case something strange happen...
            pass

    def test_submit_defaults(self):
        """Test submit_script method with its default options."""
        with stdout_redirected():
            self.cluster.submit_script(self._script, ask=False)

    def test_submit_non_default(self):
        """Test submit_script method with its non-default options."""
        launcher = "test_launcher.sh"
        self.cluster.max_walltime = "3-00:00:00"
        nb_cores_per_node = self.cluster.nb_cores_per_node // 2

        with stdout_redirected():
            self.cluster.submit_script(
                self._script,
                path_resume=self._script_resume,
                name_run="test",
                path_launching_script=launcher,
                retain_script=False,
                nb_nodes=2,
                nb_cores_per_node=nb_cores_per_node,
                nb_mpi_processes=None,
                walltime="2-23:59:59",
                nb_runs=2,
                project="2001-01-01",
                queue="queuename",
                nb_switches=2,
                max_waittime="00:10:00",
                ask=False,
                bash=False,
                email="johndoe@example.com",
                interactive=True,
            )

        if os.path.exists(launcher):
            raise ValueError(
                f"PBS launching script {launcher} was left behind"
            )


if __name__ == "__main__":
    unittest.main()



"""
Default parameters for Matplotlib figures
=========================================

.. autofunction:: set_rcparams

"""

import matplotlib.pyplot as plt


def set_rcparams(fontsize=18, for_article=True, for_beamer=False, fontsize_pad=9):
    """Set the rcparams for articles or presentation"""

    params = {
        "axes.labelsize": fontsize,
        "font.size": fontsize,
        "legend.fontsize": fontsize,
        "axes.titlesize": fontsize,
        "xtick.labelsize": fontsize - 2,
        "ytick.labelsize": fontsize - 2,
        "xtick.major.pad": fontsize_pad,
        "xtick.major.pad": fontsize_pad,
    }

    if for_article or for_beamer:
        params_tmp = {
            "text.usetex": True,
            "font.family": "serif",
            "font.serif": "Computer Modern Roman",
            "font.sans-serif": "Computer Modern Roman",
            "ps.usedistiller": "xpdf",
        }

        for k, v in list(params_tmp.items()):
            params[k] = v

    if for_beamer:
        params["font.family"] = "sans-serif"
        preamble = r"""\usepackage[cm]{sfmath}"""
        plt.rc("text.latex", preamble=preamble)

    plt.rcParams.update(params)



from __future__ import print_function

from time import sleep
import unittest

from ..daemons import DaemonThread, DaemonProcess


def func(self):
    assert self._args[0] == "hello"
    assert list(self._kwargs.keys())[0] == "word"
    while self.keepgoing.value:
        print("!", end="")
        sleep(0.01)


class DThread(DaemonThread):
    def run(self):
        func(self)


class DProcess(DaemonProcess):
    def run(self):
        func(self)


class TestDaemonds(unittest.TestCase):
    """Test fluiddyn.util.daemons module."""

    def test_daemons(self):
        dt = DThread(args=("hello",), kwargs={"word": 1})
        dt.start()
        sleep(0.02)
        dt.stop()
        dp = DProcess(args=("hello",), kwargs={"word": 1})
        dp.start()
        sleep(0.02)
        dp.stop()


if __name__ == "__main__":
    unittest.main()



"""
Test image saving and reading
=============================

"""

import unittest
import os
from itertools import chain
import numpy as np
from ..image import imread, imsave, imread_h5, imsave_h5, use_opencv


def err_msg(_format, _type, path):
    """Assertion error message. """
    return "while testing format={}, data type={} saved as {}.".format(
        _format, _type, path
    )


class TestImage(unittest.TestCase):
    """Test fluiddyn.io.image module."""

    @classmethod
    def setUpClass(cls):
        def im(maxval, dtype, shape=(10, 10)):
            """Generate a test image."""
            return (maxval * np.random.random(shape)).astype(dtype)

        cls.paths = {
            None: "test_image_none.png",
            "PNG": "test_image.png",
            #  TIFF image saving with Pillow gives error
            # 'TIFF': 'test_image.tif',
        }

        cls.paths_h5 = {"gray8": "test_gray.h5", "color8": "test_color.h5"}

        cls.images = {
            "gray8": im(2 ** 8 - 1, np.uint8),
            "gray16": im(2 ** 16 - 1, np.int32),
            "gray8f": im(2 ** 8 - 1, np.float16),
            "gray16f": im(2 ** 16 - 1, np.float32),
            "color8": im(2 ** 8 - 1, np.uint8, (10, 10, 3)),
        }

    @classmethod
    def tearDownClass(cls):
        for path in chain(cls.paths.values(), cls.paths_h5.values()):
            if os.path.exists(path):
                os.remove(path)
            if os.path.exists(path + ".h5"):
                os.remove(path + ".h5")

    def test_save_load_image(self):
        """Test save to and load from image file."""
        for _format, path in self.paths.items():
            for _type, image in self.images.items():
                if _type == "color8":
                    continue  # Cannot handle 3D images now

                err = err_msg(
                    _format, _type, path
                ) + f" Function imread from OpenCV={use_opencv}"
                as_int = bool(_type.endswith("f"))
                try:
                    imsave(path, image, _format, as_int)
                    image2 = imread(path)
                except Exception as e:
                    raise Exception(err) from e

                if as_int:
                    np.testing.assert_array_almost_equal(image, image2, 0, err)
                else:
                    np.testing.assert_array_equal(image, image2, err)

    def test_save_load_hdf5(self):
        """Test save to and load from hdf5 file."""
        _format = "HDF5"
        for _type, path in self.paths_h5.items():
            image = self.images[_type]
            imsave_h5(path, image, splitext=False)
            imsave_h5(path, image)
            image2 = imread_h5(path)
            np.testing.assert_array_equal(
                image, image2, err_msg(_format, _type, path)
            )


if __name__ == "__main__":
    unittest.main()



"""MultiFile (:mod:`fluiddyn.io.tee`)
=====================================

http://stackoverflow.com/a/16551730


"""

# bug
# from builtins import object


class MultiFile:
    """MultiFile

    Notes
    -----

    Use for example like this::

        sys.stdout = MultiFile([sys.stdout, open('myfile.txt', 'w')])

    """

    def __init__(self, files):
        self._files = files

    def __getattr__(self, attr, *args):
        return self._wrap(attr, *args)

    def _wrap(self, attr, *args):
        def g(*a, **kw):
            for f in self._files:
                res = getattr(f, attr, *args)(*a, **kw)
            return res

        return g



"""
Serie of arrays (:mod:`fluidsim.util.serieofarrays`)
====================================================

Provides classes to iterate over files.

Example::

  serie = SerieOfArraysFromFiles(path)

  def indslices_from_indserie(iserie):
      indslices = copy(serie._index_slices_all_files)
      indslices[0] = [iserie, iserie+1, 1]
      return indslices

  series = SeriesOfArrays(serie, indslices_from_indserie)

  for serie in series:
      print([name for name in serie])

API:

.. autoclass:: SerieOfArrays
   :members:
   :private-members:

.. autoclass:: SerieOfArraysFromFiles
   :members:
   :private-members:

.. autoclass:: SeriesOfArrays
   :members:
   :private-members:


"""

import os
from glob import glob, escape
from copy import copy, deepcopy
import itertools
from math import ceil, log10


# import re

try:
    import pims
except ImportError:
    pass

from fluiddyn.io import Path
from fluiddyn.io.image import imread, extensions_movies


def get_nb_arrays_in_file(fname):
    with pims.open(fname) as images:
        return images.len()


def compute_slices(str_slice):
    """Return a tuple of slices"""
    slices = []
    parts = str_slice.split(",")

    for part in parts:
        try:
            index = eval(part)
        except SyntaxError:
            parts_slice = []
            for p in part.split(":"):
                if p.strip() == "":
                    parts_slice.append(None)
                else:
                    parts_slice.append(eval(p))

            slices.append(slice(*parts_slice))
        else:
            if not isinstance(index, int):
                raise ValueError
            slices.append(index)

    return tuple(slices)


class SerieOfArrays:
    """Serie of arrays used for post-processing.

    Parameters
    ----------

    path : str
        The path of the base directory or of a file example.

    Attributes
    ----------
    path : str
        The path of the base directory.

    """

    def __init__(self, path):
        path = os.path.expanduser(path)
        if os.path.isfile(path):
            self.path_dir, self.filename_given = os.path.split(path)
        elif os.path.isdir(path):
            self.path_dir = path
            sep = os.path.sep
            filelist = glob(path + sep + "*")
            filelist.sort()
            self.filename_given = filelist[0].split(sep)[-1]
        else:
            if len(path) == 0:
                raise ValueError(
                    "The provided (empty) string does not point on any "
                    "existing path."
                )

            filelist = glob(path)
            if len(filelist) == 0:
                raise ValueError(
                    "The provided string does not point on any existing path. "
                    "The string is:\n" + path
                )

            self.path_dir, self.filename_given = os.path.split(filelist[0])

        path = os.path.join(self.path_dir, self.filename_given)
        if not os.path.isfile(path):
            raise ValueError(
                "The path given does not point towards a file "
                "but towards:\n" + path
            )

        if "." in self.filename_given:
            # bad hack
            if "[" in self.filename_given:
                self.extension_file = self.filename_given.rsplit(".", 1)[-1]
            else:
                self.extension_file = ".".join(self.filename_given.split(".")[1:])
        else:
            self.extension_file = ""


class SerieOfArraysFromFiles(SerieOfArrays):
    """Serie of arrays saved in files (images, netcdf, etc.).

    Parameters
    ----------

    path : str
        The path of the base directory or of a file example.

    index_slices : None or iterable of iterables or str

        Iterable of slides (start, stop, step).

        Can also be a string of the form "0:2:6, 1" (in this case for two
        indexes).

    Attributes
    ----------
    path_dir : str
        The path of the base directory.

    index_slices : list of list
        Lists of slides "[start, stop, step]" (one for each index).
        This list can be changed to loop over different sets of files.

    Notes
    -----

    An instance of SerieOfArraysFromFiles is an iterable and provides
    other iterables.

    Use the function :func:`set_index_slices` to specify the files
    over which iterate.

    """

    def __init__(self, path, index_slices=None, str_slices=None):

        super().__init__(path)

        self.base_name = "".join(
            itertools.takewhile(
                lambda c: not c.isdigit(),
                self.filename_given[: -(1 + len(self.extension_file))],
            )
        )

        if len(self.base_name) > 0 and not self.base_name[-1].isalpha():
            self.base_name = self.base_name[:-1]

        # remove base_name
        remains = str(self.filename_given[len(self.base_name) :])

        # remove extension
        if self.extension_file != "":
            remains = remains[: -(1 + len(self.extension_file))]

        # separator between base and index
        if len(remains) > 0 and not remains[0].isdigit():
            self._separator_base_index = remains[0]
            remains = remains[1:]
        else:
            self._separator_base_index = ""

        self._index_types = []
        self._index_lens = []
        self._index_separators = []
        while len(remains) != 0:
            if remains[0].isdigit():
                test_type = str.isdigit
                self._index_types.append("digit")
            elif remains[0].isalpha():
                test_type = str.isalpha
                self._index_types.append("alpha")
            index = "".join(itertools.takewhile(test_type, remains))
            self._index_lens.append(len(index))
            remains = remains[len(index) :]
            if len(remains) > 0:
                if not str.isalnum(remains[0]):
                    self._index_separators.append(remains[0])
                    remains = remains[1:]
                else:
                    self._index_separators.append("")

        if len(self._index_separators) < len(self._index_types):
            self._index_separators.append("")

        self.nb_indices_name_file = len(self._index_types)

        str_glob_indices = ""
        for separator in self._index_separators:
            str_glob_indices += "*" + escape(separator)

        str_glob = (
            self.base_name + escape(self._separator_base_index) + str_glob_indices
        )
        if self.extension_file != "":
            str_glob += "." + self.extension_file
        str_glob = os.path.join(self.path_dir, str_glob)
        paths = glob(str_glob)
        if not paths:
            raise ValueError(
                "There is no data in the provided path directory: " + str_glob
            )

        file_names = [os.path.basename(p) for p in paths]
        file_names.sort()

        # for files containing more than one image
        if self.extension_file in extensions_movies:
            self._from_movies = True
            self.nb_indices = self.nb_indices_name_file + 1
            if len(file_names) == 0:
                names = []
            else:
                self._nb_arrays_file = {}
                self._nb_arrays_file[
                    paths[0]
                ] = self.nb_arrays_in_one_file = get_nb_arrays_in_file(paths[0])

                self._format_index = (
                    "[{"
                    + ":0{}d".format(int(ceil(log10(self.nb_arrays_in_one_file))))
                    + "}]"
                )
                str_internal_index = [
                    self._format_index.format(i)
                    for i in range(self.nb_arrays_in_one_file)
                ]
                names = []
                for fname in file_names:
                    names.extend(fname + s for s in str_internal_index)
        else:
            self._from_movies = False
            names = file_names
            self.nb_indices = self.nb_indices_name_file

        indices_all_files = [
            self.compute_indices_from_name(name) for name in names
        ]

        indices_indices = list(zip(*indices_all_files))
        self._index_slices_all_files = []
        for i_ind in range(self.nb_indices):
            self._index_slices_all_files.append(
                [min(indices_indices[i_ind]), max(indices_indices[i_ind]) + 1, 1]
            )

        if isinstance(index_slices, str):
            self.set_index_slices_from_str(index_slices)
        elif index_slices is None:
            self._index_slices = copy(self._index_slices_all_files)
        else:
            self.set_index_slices(*index_slices)

    def set_index_slices_from_str(self, str_slices):
        """Set index_slices from a string.

        """
        slices = compute_slices(str_slices)
        index_slices = []
        for slice_ in slices:
            if isinstance(slice_, slice):
                step = slice_.step or 1
                index_slices.append((slice_.start, slice_.stop, step))
            else:
                index_slices.append(slice_)
        self.set_index_slices(*index_slices)

    def get_arrays(self):
        """Get the arrays on the serie."""
        return tuple(a for a in self.iter_arrays())

    def get_name_files(self):
        """Get the names of the files of the serie."""
        return tuple(n for n in self.iter_name_files())

    def get_name_arrays(self):
        """Get the name of the arrays of the serie."""
        return tuple(n for n in self.iter_name_arrays())

    def get_array_from_name(self, name):
        """Get the array from its name."""
        return imread(os.path.join(self.path_dir, name))

    def get_array_from_indices(self, *indices):
        """Get an array from its indices.

        Parameters
        ----------

        *indices :

          As many indices as used in the serie. For example with names of the
          form 'im100a.png', 2 indices are needed.

        """
        return self.get_array_from_name(self.compute_name_from_indices(*indices))

    def get_array_from_index(self, index):
        """Get the ith array of the serie.

        Parameters
        ----------

        index: int

          Index of the array, for example 0 to get the first array.

        """
        indices = [t for t in self.iter_indices()][index]
        return self.get_array_from_indices(*indices)

    def get_path_all_files(self):
        """Get all paths found from path_dir and base_name."""
        str_glob = os.path.join(self.path_dir, self.base_name + "*")
        paths = glob(str_glob)
        paths.sort()
        return paths

    def get_path_files(self):
        """Get all paths of the serie.

        If the serie is formed from arrays in one file, only one path is given.

        """
        return (
            os.path.join(self.path_dir, name) for name in self.get_name_files()
        )

    def get_path_arrays(self):
        """Get all paths of the arrays of the serie.

        If the serie is formed from arrays in one file, ``len(path) ==
        len(arrays)``.

        """
        return (
            os.path.join(self.path_dir, name) for name in self.get_name_arrays()
        )

    def get_name_path_arrays(self):
        return (
            (name, os.path.join(self.path_dir, name))
            for name in self.get_name_arrays()
        )

    def check_all_files_exist(self):
        """Check that all files exist."""
        name_files = self.get_name_files()
        return all([self.isfile(name) for name in name_files])

    def check_all_arrays_exist(self):
        """Check that all arrays exists."""
        if not self._from_movies:
            return self.check_all_files_exist()

        if not self.check_all_files_exist():
            return False

        for indices in self.iter_indices():
            internal_index = indices[-1]
            if internal_index >= self.nb_arrays_in_one_file:
                return False

        for path in self.get_path_files():
            if path not in self._nb_arrays_file:
                self._nb_arrays_file[path] = get_nb_arrays_in_file(path)

        if not all(
            self.nb_arrays_in_one_file == nb
            for nb in self._nb_arrays_file.values()
        ):
            return False

        return True

    def iter_indices(self):
        """Iterator on the indices.

        ``len(indices) == self.nb_indices``

        """
        lists = [list(range(*s)) for s in self._index_slices]
        for indices in itertools.product(*lists):
            yield indices

    def __len__(self):
        return sum(1 for _ in self.iter_indices())

    def iter_name_files(self):
        """Iterator on the file names."""
        names = []
        for name in self.iter_name_arrays():
            if name.endswith("]"):
                name = name.split("[")[0]
                if name not in names:
                    names.append(name)
                    yield name
            else:
                yield name

    def iter_name_arrays(self):
        """Iterator on the array names."""
        for indices in self.iter_indices():
            yield self.compute_name_from_indices(*indices)

    __iter__ = iter_name_arrays

    def iter_path_files(self):
        """Iterator on the file paths."""
        for name in self.iter_name_files():
            yield os.path.join(self.path_dir, name)

    def iter_arrays(self):
        """Iterator on the arrays of the serie."""
        for name in self.iter_name_arrays():
            yield self.get_array_from_name(name)

    def _compute_strindices_from_indices(self, *indices):
        """Compute the string corresponding to the indices.

        Parameters
        ----------

        indices: iterable of int.

        """
        if self._from_movies:
            indices = indices[:-1]

        nb_indices = len(indices)
        if nb_indices != self.nb_indices_name_file:
            raise ValueError("nb_indices != self.nb_indices")

        str_indices = ""
        for i in range(nb_indices):
            if self._index_types[i] == "digit":
                code_format = "{:0" + str(self._index_lens[i]) + "d}"
                str_index = code_format.format(indices[i])
            elif self._index_types[i] == "alpha":
                if indices[i] > 25:
                    raise ValueError('"alpha" index larger than 25.')

                str_index = chr(ord("a") + indices[i])
            else:
                raise Exception('The type should be "digit" or "alpha".')

            str_indices += str_index + self._index_separators[i]
        return str_indices

    def compute_name_from_indices(self, *indices):
        """Compute a file name from a list of indices.

        Parameters
        ----------

        indices: iterable of int.
        """
        name = (
            self.base_name
            + self._separator_base_index
            + self._compute_strindices_from_indices(*indices)
        )
        if self.extension_file != "":
            name += "." + self.extension_file

        if self._from_movies:
            name += self._format_index.format(indices[-1])

        return name

    def compute_indices_from_name(self, name):
        """Compute a list of indices from a file name.

        Parameters
        ----------

        name: str
           Name of the array.
        """
        str_indices = str(name[len(self.base_name) :])

        if self._separator_base_index != "":
            str_indices = str_indices[1:]

        if str_indices.endswith("]"):
            str_indices, internal_index = str_indices.split("[")
            internal_index = int(internal_index[:-1])
        else:
            internal_index = None

        if self.extension_file != "":
            str_indices = str_indices[: -(len(self.extension_file) + 1)]

        remains = str_indices
        indices = []
        for i_ind in range(self.nb_indices_name_file):
            if self._index_types[i_ind] == "digit":
                test_type = str.isdigit
            elif self._index_types[i_ind] == "alpha":
                test_type = str.isalpha
            else:
                raise Exception('The type should be "digit" or "alpha".')

            index = "".join(itertools.takewhile(test_type, remains))
            remains = remains[len(index) :]
            if self._index_separators[i_ind] != "":
                remains = remains[1:]

            if self._index_types[i_ind] == "digit":
                index = int(index)
            elif self._index_types[i_ind] == "alpha":
                index = ord(index) - ord("a")

            indices.append(index)

        if internal_index is not None:
            indices.append(internal_index)

        assert len(remains) == 0
        return indices

    def isfile(self, path):
        """Check whether a path or name corresponds to an existing file."""
        if not os.path.isabs(path):
            path = os.path.join(self.path_dir, path)
        return os.path.exists(path)

    def get_index_slices_all_files(self):
        """Get nb_indices "slices" (to get all the arrays in the directory).

        The "slices" are tuples of size 1 (``(start,)``), 2 (``(start, stop)``)
        or 3 (``(start, stop, step)``).

        """
        return self._index_slices_all_files

    def get_index_slices(self):
        """Get nb_indices "slices" to get all the arrays of the serie.

        The "slices" are tuples of size 1 (``(start,)``), 2 (``(start, stop)``)
        or 3 (``(start, stop, step)``).
        """
        return self._index_slices

    def set_index_slices(self, *index_slices):
        """Set the (nb_indices) "slices".

        The "slices" are tuples of size 1 (``(start,)``), 2 (``(start, stop)``)
        or 3 (``(start, stop, step)``).

        """
        index_slices = list(index_slices)
        if len(index_slices) != self.nb_indices:
            raise ValueError(
                "len(index_slices) != self.nb_indices\n"
                "filename_given = {}\n".format(self.filename_given)
                + f"path_dir = {self.path_dir}"
            )

        for i, islice in enumerate(index_slices):
            if isinstance(islice, tuple):
                index_slices[i] = list(islice)
            if isinstance(islice, int):
                index_slices[i] = [islice, islice + 1]
            elif len(islice) == 1:
                index_slices[i] = [islice[0], islice[0] + 1]

        for i, islice in enumerate(index_slices):
            for ii, index in enumerate(islice):
                if index is None:
                    index_slices[i][ii] = self._index_slices_all_files[i][ii]

        self._index_slices = index_slices

    def get_nb_arrays(self):
        """Get the number of arrays in the serie."""
        return len([i for i in self.iter_indices()])

    def get_nb_files(self):
        """Get the number of files of the serie."""
        return len(self.get_name_files())


class SeriesOfArrays:
    """Series of arrays.

    This class can be used to produce series of arrays from a
    :class:`SerieOfArrays`.

    Parameters
    ----------

    serie: SerieOfArrays or str

    indslices_from_indserie: str or function

      The function has to take an integer and to return an iterable of
      indices used to compute a file name.

    """

    def __init__(
        self,
        serie,
        indslices_from_indserie,
        ind_start=0,
        ind_stop=None,
        ind_step=1,
    ):

        serie0 = serie
        indslices_from_indserie0 = indslices_from_indserie

        if isinstance(serie, (str, Path)):
            serie = str(serie)
            serie = SerieOfArraysFromFiles(serie)
        if isinstance(serie, SerieOfArraysFromFiles):
            self.serie = serie = deepcopy(serie)
        else:
            raise ValueError("serie should be a str or a SerieOfArraysFromFiles.")

        if indslices_from_indserie is None:
            indslices_from_indserie = "i:i+1"
            for i in range(serie.nb_indices - 1):
                indslices_from_indserie += ",:"

        if isinstance(indslices_from_indserie, str):
            str_ranges = indslices_from_indserie.split(",")

            def indslices_from_indserie(i):
                indslices = []
                for str_range in str_ranges:
                    indslice = []
                    indslices.append(indslice)
                    for ii, s in enumerate(str_range.split(":")):
                        if s == "":
                            if ii == 0:
                                indslice.append(0)
                            elif ii == 1:
                                indslice.append(None)
                            elif ii > 2:
                                raise ValueError

                        else:
                            indslice.append(eval(s, {"i": i}))
                return indslices

        if indslices_from_indserie(0) == indslices_from_indserie(1):
            raise ValueError(
                "It seems that the function indslices_from_indserie "
                "does not depend on the index."
            )

        self.indslices_from_indserie = indslices_from_indserie

        if ind_stop is None:
            iserie = ind_start - ind_step
            cond = True
            while cond:
                iserie += ind_step
                serie.set_index_slices(*self.indslices_from_indserie(iserie))
                cond = serie.check_all_arrays_exist()
            iserie -= 1
        else:
            if len(range(ind_start, ind_stop, ind_step)) == 0:
                raise ValueError("len(range(ind_start, ind_stop, ind_step)) == 0")

            for iserie in range(ind_start, ind_stop, ind_step):
                serie.set_index_slices(*self.indslices_from_indserie(iserie))
                if not serie.check_all_arrays_exist():
                    break

        ind_stop = iserie + 1

        self.nb_series = len(list(range(ind_start, ind_stop, ind_step)))
        self.iserie = ind_start
        self.ind_start = ind_start
        self.ind_stop = ind_stop
        self.ind_step = ind_step

        serie.set_index_slices(*self.indslices_from_indserie(self.iserie))
        _print_warning = False
        if not serie.check_all_arrays_exist():
            _print_warning = True
            print(
                "warning: this SeriesOfArrays has been initialized with "
                "parameters such that the first serie if not complete."
            )

        if self.nb_series == 0:
            _print_warning = True
            print(
                "warning: this SeriesOfArrays has been initialized with "
                "parameters such that no serie of images has been found."
            )

        if _print_warning:
            print(
                "serie={},\nindslices_from_indserie={}, ".format(
                    serie0, indslices_from_indserie0
                )
                + "ind_start={}, ind_stop={}, ind_step={}.".format(
                    ind_start, ind_stop, ind_step
                )
            )

    def __iter__(self):
        if hasattr(self, "index_series"):
            index_series = self.index_series
        else:
            index_series = list(
                range(self.ind_start, self.ind_stop, self.ind_step)
            )

        for iserie in index_series:
            self.serie.set_index_slices(*self.indslices_from_indserie(iserie))
            yield self.serie

    def items(self):
        if hasattr(self, "index_series"):
            index_series = self.index_series
        else:
            index_series = list(
                range(self.ind_start, self.ind_stop, self.ind_step)
            )

        for iserie in index_series:
            self.serie.set_index_slices(*self.indslices_from_indserie(iserie))
            yield iserie, self.serie

    def __len__(self):
        return len([s for s in self])

    def set_index_series(self, index_series):
        """Set the indices corresponding to the series.

        Parameters
        ----------

        index_series : sequence of int

        """
        self.index_series = index_series

    def get_next_serie(self):
        """Get the next serie."""
        if self.iserie < self.ind_stop:
            self.serie.set_index_slices(
                *self.indslices_from_indserie(self.iserie)
            )
            self.iserie += 1
            return self.serie

    def get_serie_from_index(self, index):
        """Get a serie from an index."""
        self.serie.set_index_slices(*self.indslices_from_indserie(index))
        return self.serie

    def get_name_all_files(self):
        """Get all file names."""
        names_all = []
        for serie in self:
            names = serie.get_name_files()
            for name in names:
                if name not in names_all:
                    names_all.append(name)
        return names_all

    def get_name_all_arrays(self):
        """Get all array names."""
        names_all = []
        for serie in self:
            names = serie.get_name_arrays()
            for name in names:
                if name not in names_all:
                    names_all.append(name)
        return names_all


# if __name__ == "__main__":

# path = (
#     os.environ['HOME'] +
#     '/Dev/Matlab/Demo/UVMAT_DEMO04_PIV_challenge_2005_CaseC/'
#     'Images'
#     '/c001a.png'
# )

# serie = SerieOfArraysFromFiles(path)

# def indslices_from_indserie(iserie):
#     indslices = copy(serie._index_slices_all_files)
#     indslices[0] = [iserie, iserie+1]
#     return indslices

# series = SeriesOfArrays(serie, indslices_from_indserie)

# for serie in series:
#     print([name for name in serie])

# print('\nOther test')

# def indslices_from_indserie(iserie):
#     indslices = copy(serie._index_slices_all_files)
#     indslices[0] = [iserie, iserie+2, 1]
#     indslices[1] = [1]
#     return indslices

# series = SeriesOfArrays(serie, indslices_from_indserie)

# for serie in series:
#     print([name for name in serie])

# path = os.environ["HOME"] + "/Dev/howtopiv/samples/Karman"

# serie = SerieOfArraysFromFiles(path, base_name="PIVlab_Karman")

# def indslices_from_indserie(iserie):
#     indslices = copy(serie._index_slices_all_files)
#     indslices[0] = [iserie + 1, iserie + 2]
#     return indslices

# series = SeriesOfArrays(serie, indslices_from_indserie)

# for serie in series:
#     print([name for name in serie])



"""
IO for multitiff files (:mod:`fluiddyn.io.multitiff`)
=====================================================

.. autofunction:: reorganize_single_frame_3Dscannedpiv_data
.. autofunction:: reorganize_piv3dscanning_doubleframe
.. autofunction:: reorganize_piv2d_singleframe
.. autofunction:: reorganize_piv2d_doubleframe

"""
from __future__ import print_function, division

from builtins import range
import sys
import time
import os
import glob
import io
from math import ceil, log10


try:
    import PIL
    from PIL import Image
except ImportError:
    pass

from .query import query_yes_no
from .image import _image_from_array


def glob_sorted(s):
    l = glob.glob(s)
    l.sort()
    return l


def _should_we_stop():
    if query_yes_no("Some folders already exist\n Do you want to continue?"):
        return False

    return True


def _save_new_file(im, base_path, outputext, erase=False):
    """Convert and save file if destination path does not exist.

    FIXME: The `im.convert` function calls result in ResourceWarning
    in Python 3.x. Almost certainly a Pillow bug. Can be suppressed by
    adding the following lines.

    >>> import warnings
    >>> warnings.simplefilter('ignore', ResourceWarning)

    """
    path_save = base_path + "." + outputext
    if not erase and os.path.exists(path_save):
        return False

    if outputext == "jp2":
        with im.convert(mode="L") as new_im:
            new_im.save(path_save, quality_mode="rate", quality_layers=[8])
    else:
        with im.convert(mode="I") as new_im:
            new_im.save(path_save)

    return True


def imsave(path, arrays, as_int=False):
    """Save a multi-frame image sequence.

    Parameters
    ----------
    path : str
        Output file name or full path.
    arrays : array-like
        A iterable containing multiple 2D numpy arrays, representing frames.
    as_int : bool
        Convert to integer or not.

    """
    if PIL.__version__ < "3.4.0":
        raise ImportError("imsave for multiframe TIFF not supported.")

    im_list = [_image_from_array(a, as_int) for a in arrays]

    if PIL.__version__ < "4.2.0":
        from warnings import warn

        warn(
            "imsave for multiframe TIFF uses an intermediate GIF workaround.",
            DeprecationWarning,
        )
        gif = im_list.pop(0)

        with io.BytesIO() as output:
            gif.save(output, format="GIF", save_all=True, append_images=im_list)
            gif.close()

            with Image.open(output) as im:
                im.save(path, format="TIFF", save_all=True)
    else:
        im_list[0].save(
            path,
            compression="tiff_deflate",
            save_all=True,
            append_images=im_list[1:],
        )


def reorganize_single_frame_3Dscannedpiv_data(
    files, nb_levels, outputdir=".", outputext="tif", erase=False
):
    """
    Reorganize data from multi tiff into a folders (one for each level).

    Parameters
    ----------

    files : str
      Root of the names of files: example files = 'van_karman_flow*'

    nb_levels : int
      Number of levels in the scanned PIV

    outputdir : '.'
      Output folder.

    outputext : str
      The output files extension

    erase : {False, bool}
      If erase, the existing files are replaced.

    Notes
    -----

    Data is organize as:
    - outputdir/level1/im0.tif, outputdir/level1/im1.tif ...
    - outputdir/level2/im0.tif, outputdir/level2/im1.tif ...

    """

    path_files = glob_sorted(files)
    if len(path_files) == 0:
        return

    nb_images = get_approximate_number_images(path_files)
    if nb_images == 0:
        return

    format_index = ":0{}d".format(int(ceil(log10(nb_images))))
    format_level = ":0{}d".format(int(ceil(log10(nb_levels))))

    index_im = 0

    for ind in range(nb_levels):
        dir_lev = (outputdir + "/level{" + format_level + "}").format(ind)
        if not os.path.exists(dir_lev):
            os.makedirs(dir_lev)

    for path_tiff in path_files:
        t_start = time.time()
        print(
            "Convert and save file\n" + path_tiff + "\nin directory\n" + outputdir
        )
        with Image.open(path_tiff) as im:
            index_im_in_tiff = 0
            while True:
                try:
                    im.seek(index_im_in_tiff)
                except EOFError:
                    break

                else:
                    base_path = (
                        outputdir
                        + "/level{"
                        + format_level
                        + "}/im{"
                        + format_index
                        + "}"
                    ).format(index_im % nb_levels, index_im // nb_levels)
                    if _save_new_file(im, base_path, outputext, erase):
                        print(
                            "\r file {}; in {:.2f} s".format(
                                index_im, time.time() - t_start
                            ),
                            end="",
                        )
                        sys.stdout.flush()

                    index_im += 1
                    index_im_in_tiff += 1
            print("")


def reorganize_piv3dscanning_doubleframe(
    files, nb_levels, outputdir=".", outputext="tif", erase=False
):
    """
    Reorganize data from multi tiff into a folders (one for each level).

    Parameters
    ----------

    files : str
      Root of the names of files: example files = 'van_karman_flow*'

    nb_levels : int
      Number of levels in the scanned PIV

    outputdir : '.'
      Output folder.

    outputext : str
      The output files extension

    erase : {False, bool}
      If erase, the existing files are replaced.

    Notes
    -----

    Data is organized as:
    - outputdir/level1/im0.tif, outputdir/level1/im1.tif ...
    - outputdir/level2/im0.tif, outputdir/level2/im1.tif ...

    """
    path_files = glob_sorted(files)
    if len(path_files) == 0:
        return

    nb_images = get_approximate_number_images(path_files)
    if nb_images == 0:
        return

    format_index = ":0{}d".format(int(ceil(log10(nb_images))))
    format_level = ":0{}d".format(int(ceil(log10(nb_levels))))

    index_im = 0

    for ind in range(nb_levels):
        dir_lev = (outputdir + "/level{" + format_level + "}").format(ind)
        if not os.path.exists(dir_lev):
            os.makedirs(dir_lev)

    for path_tiff in path_files:
        t_start = time.time()
        print(
            "Convert and save file\n" + path_tiff + "\nin directory\n" + outputdir
        )
        with Image.open(path_tiff) as im:
            index_im_in_tiff = 0
            while True:
                try:
                    im.seek(index_im_in_tiff)
                except EOFError:
                    break

                except SyntaxError:
                    print(
                        "SyntaxError with file",
                        path_tiff,
                        "\nStop the conversion for this file.",
                    )
                    return

                else:
                    index_time = index_im // nb_levels
                    if index_time % 2 == 0:
                        letter = "a"
                    else:
                        letter = "b"
                    base_path = (
                        outputdir
                        + "/level{"
                        + format_level
                        + "}/im{"
                        + format_index
                        + "}"
                        + letter
                    ).format(index_im % nb_levels, index_time // 2)
                    if _save_new_file(im, base_path, outputext, erase):
                        print(
                            "\r file {}; in {:.2f} s".format(
                                index_im, time.time() - t_start
                            ),
                            end="",
                        )
                        sys.stdout.flush()

                    index_im += 1
                    index_im_in_tiff += 1
            print("")


def count_number_images(path_files):
    # warning: can be very slow
    nb_images = 0
    for path_tiff in path_files:
        with Image.open(path_tiff) as image:
            nb_images += image.n_frames
            print(
                "taking in account file {}, nb_images = {}".format(
                    path_tiff, nb_images
                )
            )

    return nb_images


def get_approximate_number_images(path_files):

    # can be much faster than count_number_images
    path = path_files[0]
    with Image.open(path) as image:
        nb_images = image.n_frames

    return len(path_files) * nb_images


def reorganize_piv2d_singleframe(
    files, outputdir=".", outputext="tif", erase=False
):
    """
    Reorganize data from multi tiff (single frame 2D).

    Parameters
    ----------

    files : str
      Root of the names of files: example files = 'van_karman_flow*'

    outputdir : '.'
      Output folder

    outputext : str
      The output files extension

    erase : {False, bool}
      If erase, the existing files are replaced.

    Notes
    -----

    Data is organize as:
    outputdir/im0.tif, outputdir/im1.tif ...

    """
    path_files = glob_sorted(files)
    if len(path_files) == 0:
        return

    nb_images = get_approximate_number_images(path_files)
    if nb_images == 0:
        return

    format_index = ":0{}d".format(int(ceil(log10(nb_images))))

    if not os.path.exists(outputdir):
        os.makedirs(outputdir)

    index_im = 0
    for path_tiff in path_files:
        t_start = time.time()
        print(
            "Convert and save file\n" + path_tiff + "\nin directory\n" + outputdir
        )
        with Image.open(path_tiff) as im:
            index_im_in_tiff = 0
            while True:
                try:
                    im.seek(index_im_in_tiff)
                except EOFError:
                    break

                else:
                    base_path = outputdir + ("/im{" + format_index + "}").format(
                        index_im
                    )
                    if _save_new_file(im, base_path, outputext, erase):
                        print(
                            "\r file {}; in {:.2f} s".format(
                                index_im, time.time() - t_start
                            ),
                            end="",
                        )
                        sys.stdout.flush()
                    index_im += 1
                    index_im_in_tiff += 1

            print("")

        print(
            "End convert file {} in {} s".format(
                os.path.split(path_tiff)[0], time.time() - t_start
            )
        )
        sys.stdout.flush()


def reorganize_piv2d_doubleframe(
    files, outputdir=".", outputext="tif", erase=False
):
    """
    Reorganize data from multi tiff (double frame 2D).

    Parameters
    ----------
    files : str
      Root of the names of files: example files = 'van_karman_flow*'

    outputdir : '.'
      Output folder

    outputext : str
      The output files extension

    erase : {False, bool}
      If erase, the existing files are replaced.

    Notes
    -----

    Data is organize as:
    - outputdir/im0a.tif, outputdir/im0b.tif, outputdir/im1a.tif ...

    """
    path_files = glob_sorted(files)
    if len(path_files) == 0:
        return

    nb_images = get_approximate_number_images(path_files)
    if nb_images == 0:
        return

    format_index = ":0{}d".format(int(ceil(log10(nb_images))))

    index_im = 0
    if not os.path.exists(outputdir):
        os.makedirs(outputdir)

    for path_tiff in path_files:
        t_start = time.time()
        print(
            "Convert and save file\n" + path_tiff + "\nin directory\n" + outputdir
        )
        with Image.open(path_tiff) as im:
            index_im_in_tiff = 0
            while True:
                try:
                    im.seek(index_im_in_tiff)
                except EOFError:
                    break

                else:
                    if index_im % 2 == 0:
                        letter = "a"
                    else:
                        letter = "b"
                    base_path = (
                        outputdir
                        + ("/im{" + format_index + "}").format(index_im // 2)
                        + letter
                    )
                    if _save_new_file(im, base_path, outputext, erase):
                        print(
                            "\r file {}; in {:.2f} s".format(
                                index_im, time.time() - t_start
                            ),
                            end="",
                        )
                        sys.stdout.flush()
                    index_im += 1
                    index_im_in_tiff += 1

            print("")

        print(
            "End convert file {} in {} s".format(
                os.path.split(path_tiff)[0], time.time() - t_start
            )
        )
        sys.stdout.flush()



"""
Test mpi module
===============

"""

import unittest
import os
from shutil import rmtree
import io

import numpy as np

from ..mpi import _mpi_type, nb_proc, rank, printby0
from ...io.redirect_stdout import stdout_redirected


class TestMPI(unittest.TestCase):
    """Test fluiddyn.util.mpi module."""

    @classmethod
    def barrier(cls):
        if _mpi_type is not None:
            from ..mpi import comm

            comm.barrier()

    @classmethod
    def setUpClass(cls):
        cls._work_dir = "test_fluiddyn_util_mpi"
        if rank == 0:
            if not os.path.exists(cls._work_dir):
                os.mkdir(cls._work_dir)

        cls.barrier()
        os.chdir(cls._work_dir)

    @classmethod
    def tearDownClass(cls):
        os.chdir("..")
        cls.barrier()
        if rank == 0:
            rmtree(cls._work_dir)

    def test_print(self):
        """Test fluiddyn.util.mpi.printby0 to write to a buffer."""

        msg_write = "test by rank ="
        output = io.StringIO()
        with stdout_redirected(to=output):
            printby0(msg_write, rank, end="")

        output.flush()
        self.barrier()
        msg_read = output.getvalue()
        output.close()

        msg_expected = msg_write + " 0" if rank == 0 else ""
        self.assertEqual(msg_read, msg_expected)

    @unittest.skipIf(nb_proc == 1, "Meant for testing if mpi4py works.")
    def test_scatter_gather(self):
        """Test MPI Scatter and Gather functions for numpy objects."""

        from ..mpi import comm, MPI

        N_loc = 4
        N = N_loc * nb_proc
        if comm.rank == 0:
            arr = np.arange(N, dtype=np.float64)
        else:
            arr = np.empty(N, dtype=np.float64)

        arr_loc = np.empty(N_loc, dtype=np.float64)
        comm.Scatter([arr, MPI.DOUBLE], [arr_loc, MPI.DOUBLE])

        arr_loc *= 2
        if comm.rank == 0:
            arr2 = np.empty(N, dtype=np.float64)
        else:
            arr2 = None

        comm.Gather([arr_loc, MPI.DOUBLE], [arr2, MPI.DOUBLE])

        if rank == 0:
            np.testing.assert_array_equal(arr * 2, arr2)


if __name__ == "__main__":
    unittest.main()



"""IO for text files (:mod:`fluiddyn.io.txt`)
=============================================

With recent version of Numpy, it is most of the time better to use np.loadtxt
and np.savetxt...


"""


from __future__ import division, print_function

from builtins import range
import os
import numpy as np

from . import query


def num_from_str(s):
    """Return a number from a string."""
    try:
        return int(s)

    except ValueError:
        try:
            return float(s)

        except ValueError:
            raise ValueError(
                'Can not convert the string "' + s + '" into a numeric value.'
            )


def quantities_from_txt_file(name_file):
    """Read a txt file and return the numerical data."""
    if os.path.isabs(name_file):
        path_file = name_file
    else:
        path_file = os.path.abspath(name_file)

    if not os.path.exists(path_file):
        raise ValueError("file does not exist? path_file :\n" + path_file)

    f = open(path_file, "r")
    lines = f.readlines()
    f.close()

    lines_with_num_data = []
    nb_nums_per_line = 0

    while len(lines) > 0:
        words = lines[0].split()
        del (lines[0])
        try:
            float(words[0])
            lines_with_num_data.append(words)
            if nb_nums_per_line < len(words):
                nb_nums_per_line = len(words)
        except (ValueError, IndexError):
            pass

    # create a list of lists
    lol = [[] for l in range(nb_nums_per_line)]

    for words in lines_with_num_data:
        for iw, word in enumerate(words):
            lol[iw].append(num_from_str(word))

    for iq in range(nb_nums_per_line):
        lol[iq] = np.array(lol[iq])

    return tuple(lol)


def save_quantities_in_txt_file(name_file, args, erase=False):
    """Save array_like objects in a text file."""
    try:
        args[0][0]
    except ValueError:
        raise ValueError("args has to be 2 times iterable...")

    nb_quantities = len(args)

    if nb_quantities < 1:
        raise ValueError("At least one quantity has to be given.")

    nb_values_1q = len(args[0])
    for iq in range(1, nb_quantities):
        if len(args[iq]) != nb_values_1q:
            raise ValueError("All quantities should have the same length.")

    if os.path.isfile(name_file) and not erase:
        answer = query.query(
            "The file "
            + name_file
            + " already exists. "
            + "What do you want to do?\n"
            + "[Don't save / erase file and save / modify name] ",
            default="Don't save",
        )
        if answer.lower().startswith("do"):
            print("The quantities have not been saved.")
            return

        elif answer.lower().startswith("e"):
            print("Existing file will be erased.")
        else:
            # need to be improved !
            print(
                "Sorry. This option has to be implemented. "
                "The quantities have not been saved."
            )
            return

    with open(name_file, "w") as f:
        for iv in range(nb_values_1q):
            for iq in range(nb_quantities):
                f.write(repr(args[iq][iv]))
                f.write(" ")
            f.write("\n")


if __name__ == "__main__":

    t = quantities_from_txt_file("files.py")
    print(t)

    save_quantities_in_txt_file("test", t)


"""
zekfzlnfk

zklflzefk

1   3. 1

2   5. 4.
3  4  9
"""



"""
Test txt functions
==================

"""

import unittest
import os
from shutil import rmtree

from ..txt import quantities_from_txt_file, save_quantities_in_txt_file

txt_example = """
zekfzlnfk

zklflzefk

1  3. 1

2   5.
3  4.  9
"""

t_example = ([1, 2, 3], [3.0, 5.0, 4.0], [1, 9])


class TestTxt(unittest.TestCase):
    """Test fluiddyn.io.txt module."""

    def setUp(self):
        self._work_dir = "test_fluiddyn_io_txt"
        if not os.path.exists(self._work_dir):
            os.mkdir(self._work_dir)

        os.chdir(self._work_dir)

        self.path_in = "path_in"

        with open(self.path_in, "w") as f:
            f.write(txt_example)

    def tearDown(self):
        os.chdir("..")
        rmtree(self._work_dir)

    def test_txt(self):

        t = quantities_from_txt_file(self.path_in)

        for i, l_example in enumerate(t_example):
            for j, a in enumerate(l_example):
                self.assertEqual(a, t[i][j])

        t = ([1, 2, 3],) * 3

        save_quantities_in_txt_file("path_out", t)


if __name__ == "__main__":
    unittest.main()
